import pathlib
import pytest

pathDataSamples = pathlib.Path(__file__).parent / "dataSamples"

listOfFiles = ["pink-20RMS60sec.wav",
"pink-40RMS60sec.wav",
"pink-60RMS60sec.wav",
"testParkMono96kHz32float12.1sec.wav",
"testPink2ch7.1sec.wav",
"testSine2ch5sec.wav",
"testSine2ch5secCopy1.wav",
"testTrain2ch48kHz6.3sec.wav",
"testVideo11sec.mkv",
"testWooWooMono16kHz32integerClipping9sec.wav",
]
